!###############################################################################
!                                                                              #
!         Objective:      Read a matrix of expression values, compute          #
!                         correlation coexpression and assign distribution     #
!                         to one of the 8 "template" distributions             #
!                                                                              #
!         Authors:        Antonio Reverter (Tony.Reverter-Gomez@csiro.au)      #
!                         Pamela Alexandere (Pamela.Alexandre@csiro.au)        #
!                                                                              #
!         Last Revision:  July 2019                                            #
!                                                                              #
!###############################################################################
PROGRAM CoExprDistrib

IMPLICIT NONE

INTEGER, PARAMETER :: nshapes = 8, &   ! Number of Template shapes
                      nbins = 8        ! Number of bins within shape

INTEGER  :: ngenes, &      ! Number of genes
            nvals, &       ! Number of correlation values (ngenes-1)
            nconditions    ! Experimenal conditions = dimension
                           ! of gene expression vectors

TYPE gene_record
   CHARACTER(LEN=100) :: id
   double precision,allocatable :: expr(:)
! bin proportions, distances and probabilities
   double precision :: bp(nbins), dist(nshapes), prob(nshapes)
   double precision :: mn, sd, sk, ku    ! stats on correl coexpr for each gene
   integer :: shape(1),    & ! Allocated shaped based on highest probability
              alternative, & ! 0/1 flag if in top 5% of non-null (ie. 1,2 vs 3 to 8)
              nmodes         ! 0/1/2 flag is in 90% intermediate, 5% unimodal, 5% bimodal
END TYPE gene_record
TYPE(gene_record),allocatable :: gene(:)

TYPE shape_template
   double precision :: bp(nbins)        ! bin proportion
   double precision :: mn,sd,sk,ku  ! mean, sd, skewness, kurtosis
END TYPE shape_template
TYPE(shape_template) :: sh(nshapes)

double precision, allocatable :: gene_corr(:,:), now(:), &
                                 probdiff(:)  ! probability difference for testing
                                              ! alternative and nmodes tests

double precision :: corr,mindist,maxdist,range,sumprob, &
                    pct5_threshold, & ! the 5% threshold for testing
                    pct95_threshold   ! the 95% threshold for testing n modes

INTEGER :: i,j,k,io,maxlengthid
CHARACTER(LEN=100) :: infile, line
CHARACTER(LEN=100) :: outfile1, outfile2, outfile3, outfile4, outfile5
CHARACTER(LEN=100) :: fmat1, fmat2, fmat3, fmat4, fmat5

! NOTE: 5 output files for correlations, bin proportions, &
!       distances and probabilities to each shape, and tests results

write(*,*) 'Name of space-delimited file with Log2 normalised expressions for each gene across all samples/conditions'
read(*,*)infile

write(*,*)' '
outfile1=trim(adjustl(infile))//'_Corrs.txt'
write(*,*) 'The output file with correlations is ',trim(outfile1)
outfile2=trim(adjustl(infile))//'_Bins.txt'
write(*,*) 'The output file with bin proportions is ',trim(outfile2)
outfile3=trim(adjustl(infile))//'_Dist.txt'
write(*,*) 'The output file with Distances is ',trim(outfile3)
outfile4=trim(adjustl(infile))//'_Prob.txt'
write(*,*) 'The output file with Distances is ',trim(outfile4)
outfile5=trim(adjustl(infile))//'_Results.txt'
write(*,*) 'The output file with Test Results is ',trim(outfile5)

OPEN(20,file=infile,status='old',form='formatted')

OPEN(50,file=outfile1,status='replace',form='formatted')
OPEN(60,file=outfile2,status='replace',form='formatted')
OPEN(70,file=outfile3,status='replace',form='formatted')
OPEN(80,file=outfile4,status='replace',form='formatted')
OPEN(90,file=outfile5,status='replace',form='formatted')
!#################################################################

ngenes=0
do
  read(20,'(a1)',iostat=io)line(1:1)
  if(io/=0) exit
  ngenes=ngenes+1
enddo
rewind(20)
write(*,*)'number of genes= ',ngenes
call number_of_fields(20,nconditions)
nconditions=nconditions-1
write(*,*) 'number of conditions or samples: ',nconditions
IF( nconditions < 5 )THEN
   write(*,*)'Sorry mate, you need at least 5 conditions to run this'
   write(*,*)'Bye'
   STOP
ENDIF

allocate(gene_corr(ngenes,ngenes),gene(ngenes),probdiff(ngenes))
gene_corr = 1.d-4
do i=1,ngenes
    allocate(gene(i)%expr(nconditions))
enddo

maxlengthid=0
DO i = 1,ngenes
   READ(20,*)gene(i)%id,gene(i)%expr(1:nconditions)
   maxlengthid=maxval((/maxlengthid,len_trim(adjustl(gene(i)%id))/))
ENDDO
print *,'max length of gene name: ',maxlengthid

!###################################################
! Process Gene Correlations
!###################################################
! create format for output1
write(fmat1,'(a,i0,a)') '(2a',maxlengthid+1,',1x,f10.5)'
DO i = 1, ngenes-1
   if(mod(i,250)==0)write(*,*)'Correlations computed for gene',i
   DO j = i+1, ngenes
      gene_corr(i,j) = corr(gene(i)%expr,gene(j)%expr,nconditions)
      gene_corr(j,i) = gene_corr(i,j)
      WRITE(50,fmat1)gene(i)%id,gene(j)%id,gene_corr(i,j)
   ENDDO
ENDDO

!#############################################################
! Process Bin Proportions, mean, sd, skewness and kurtosis
!#############################################################
!#############################################################
! Process Distances to each of the 8 template shapes
!#############################################################
! create format for output2 and output3
write(fmat2,'(a,i0,a)') '(1a',maxlengthid+1,',1x,8f8.5,4f12.5)'
write(fmat3,'(a,i0,a)') '(1a',maxlengthid+1,',1x,8f10.5)'
write(fmat4,'(a,i0,a)') '(1a',maxlengthid+1,',1x,8f10.5)'

! Initialise shape templates
sh(1)%bp(1)=0.05; sh(2)%bp(1)=0.02; sh(3)%bp(1)=0.15; sh(4)%bp(1)=0.01
sh(1)%bp(2)=0.10; sh(2)%bp(2)=0.05; sh(3)%bp(2)=0.30; sh(4)%bp(2)=0.02
sh(1)%bp(3)=0.15; sh(2)%bp(3)=0.15; sh(3)%bp(3)=0.25; sh(4)%bp(3)=0.03
sh(1)%bp(4)=0.20; sh(2)%bp(4)=0.28; sh(3)%bp(4)=0.15; sh(4)%bp(4)=0.09
sh(1)%bp(5)=0.20; sh(2)%bp(5)=0.28; sh(3)%bp(5)=0.09; sh(4)%bp(5)=0.15
sh(1)%bp(6)=0.15; sh(2)%bp(6)=0.15; sh(3)%bp(6)=0.03; sh(4)%bp(6)=0.25
sh(1)%bp(7)=0.10; sh(2)%bp(7)=0.05; sh(3)%bp(7)=0.02; sh(4)%bp(7)=0.30
sh(1)%bp(8)=0.05; sh(2)%bp(8)=0.02; sh(3)%bp(8)=0.01; sh(4)%bp(8)=0.15

sh(5)%bp(1)=0.28; sh(6)%bp(1)=0.05; sh(7)%bp(1)=0.05; sh(8)%bp(1)=0.05
sh(5)%bp(2)=0.15; sh(6)%bp(2)=0.25; sh(7)%bp(2)=0.30; sh(8)%bp(2)=0.15
sh(5)%bp(3)=0.05; sh(6)%bp(3)=0.15; sh(7)%bp(3)=0.20; sh(8)%bp(3)=0.10
sh(5)%bp(4)=0.02; sh(6)%bp(4)=0.05; sh(7)%bp(4)=0.10; sh(8)%bp(4)=0.05
sh(5)%bp(5)=0.02; sh(6)%bp(5)=0.05; sh(7)%bp(5)=0.05; sh(8)%bp(5)=0.10
sh(5)%bp(6)=0.05; sh(6)%bp(6)=0.15; sh(7)%bp(6)=0.10; sh(8)%bp(6)=0.20
sh(5)%bp(7)=0.15; sh(6)%bp(7)=0.25; sh(7)%bp(7)=0.15; sh(8)%bp(7)=0.30
sh(5)%bp(8)=0.28; sh(6)%bp(8)=0.05; sh(7)%bp(8)=0.05; sh(8)%bp(8)=0.05

sh(1)%mn=0d0;    sh(1)%sd=0.456;  sh(1)%sk=0d0;    sh(1)%ku=-0.660
sh(2)%mn=0d0;    sh(2)%sd=0.354;  sh(2)%sk=0d0;    sh(2)%ku=0.065
sh(3)%mn=-0.387; sh(3)%sd=0.382;  sh(3)%sk=0.857;  sh(3)%ku=0.640
sh(4)%mn=0.387;  sh(4)%sd=0.382;  sh(4)%sk=-0.857; sh(4)%ku=0.640
sh(5)%mn=0d0;    sh(5)%sd=0.752;  sh(5)%sk=0d0;    sh(5)%ku=-1.771
sh(6)%mn=0d0;    sh(6)%sd=0.566;  sh(6)%sk=0d0;    sh(6)%ku=-1.535
sh(7)%mn=-0.137; sh(7)%sd=0.532;  sh(7)%sk=0.500;  sh(7)%ku=-1.111
sh(8)%mn=0.137;  sh(8)%sd=0.532;  sh(8)%sk=-0.500; sh(8)%ku=-1.111

nvals = ngenes-1
allocate(now(nvals))
DO i = 1, ngenes
   k = 0
   DO j = 1, ngenes
      IF( i .NE. j )THEN
         k = k + 1
         now(k) = gene_corr(i,j)
      ENDIF
   ENDDO

   gene(i)%bp(1) = 1d0*count(now<-0.75d0)/nvals
   gene(i)%bp(2) = 1d0*count(now<-0.50d0)/nvals - gene(i)%bp(1)
   gene(i)%bp(3) = 1d0*count(now<-0.25d0)/nvals - sum(gene(i)%bp(1:2))
   gene(i)%bp(4) = 1d0*count(now<0.0d0)/nvals - sum(gene(i)%bp(1:3))
   gene(i)%bp(5) = 1d0*count(now<0.25d0)/nvals - sum(gene(i)%bp(1:4))
   gene(i)%bp(6) = 1d0*count(now<0.50d0)/nvals - sum(gene(i)%bp(1:5))
   gene(i)%bp(7) = 1d0*count(now<0.75d0)/nvals - sum(gene(i)%bp(1:6))
   gene(i)%bp(8) = 1d0 - sum(gene(i)%bp(1:7))

   gene(i)%mn = mean(now)
   gene(i)%sd = sd(now)
   gene(i)%sk = (1d0/nvals)*sum( ( (now-gene(i)%mn)/gene(i)%sd )**3 )
   gene(i)%ku = (1d0/nvals)*sum( ( (now-gene(i)%mn)/gene(i)%sd )**4 ) - 3.

   WRITE(60,fmat2)gene(i)%id,  &
                  gene(i)%bp(1),gene(i)%bp(2),gene(i)%bp(3),gene(i)%bp(4), &
                  gene(i)%bp(5),gene(i)%bp(6),gene(i)%bp(7),gene(i)%bp(8), &
                  gene(i)%mn,gene(i)%sd,gene(i)%sk,gene(i)%ku

   ! Compute Euclidean distances across the 8 bins plus 4 point estimates

   gene(i)%dist = 0d0
   DO j = 1, nshapes
      DO k = 1, nbins
         gene(i)%dist(j) = gene(i)%dist(j) + &
                           (gene(i)%bp(k) - sh(j)%bp(k))**2
      ENDDO
      gene(i)%dist(j) = gene(i)%dist(j) + (gene(i)%mn-sh(j)%mn)**2 + &
                                          (gene(i)%sd-sh(j)%sd)**2 + &
                                          (gene(i)%sk-sh(j)%sk)**2 + &
                                          (gene(i)%ku-sh(j)%ku)**2
      gene(i)%dist(j) = sqrt( gene(i)%dist(j) )
   ENDDO

ENDDO

DO i = 1, ngenes
   WRITE(70,fmat3)gene(i)%id,gene(i)%dist(1:nshapes)
ENDDO
   
DO i = 1, ngenes
   mindist=minval(gene(i)%dist)
   maxdist=maxval(gene(i)%dist)
   range = maxdist - mindist
   DO j = 1, nshapes  ! Similarities on a 0,1 scale
      gene(i)%prob(j)=1d0-(gene(i)%dist(j)-mindist)/range
   ENDDO
   sumprob=sum(gene(i)%prob)
   DO j = 1, nshapes  ! Probabilities adding to 100% across all shapes
      gene(i)%prob(j)=100d0*gene(i)%prob(j)/sumprob
   ENDDO
   WRITE(80,fmat4)gene(i)%id,gene(i)%prob(1:nshapes)
ENDDO

!#######################################################################
! Final file with allocated shape and test of alternative and n modes
!#######################################################################
!write(fmat5,'(a,i0,a)') '(1a',maxlengthid+1,',1x,3i3)'
write(fmat5,'(a,i0,a)') '(1a',maxlengthid+1,',1x,1i3)'

! Find the shape of maximum probability
!##################################################################
DO i = 1, ngenes
   gene(i)%shape = MAXLOC(gene(i)%prob(1:nshapes))
ENDDO

! Test for Alternative distribution shape (ie. other than 1 or 2)
!##################################################################
DO i = 1, ngenes
   probdiff(i) = mean(gene(i)%prob(1:2)) - mean(gene(i)%prob(3:8))
ENDDO
CALL quicksort(probdiff,1,ngenes)
pct5_threshold = probdiff(int(ngenes*0.05))
DO i = 1, ngenes
   probdiff(i) = mean(gene(i)%prob(1:2)) - mean(gene(i)%prob(3:8))
ENDDO
gene%alternative = 0
DO i = 1, ngenes
   IF (probdiff(i) < pct5_threshold ) gene(i)%alternative = 1
ENDDO

! Test for uni- vs bi-modal shape (ie. 0/1/2 for undecided, 1 or 2 modes)
!#########################################################################
DO i = 1, ngenes
   probdiff(i) = mean(gene(i)%prob(1:4)) - mean(gene(i)%prob(5:8))
ENDDO
CALL quicksort(probdiff,1,ngenes)
pct5_threshold = probdiff(int(ngenes*0.05))
pct95_threshold = probdiff(int(ngenes*0.95))
DO i = 1, ngenes
   probdiff(i) = mean(gene(i)%prob(1:4)) - mean(gene(i)%prob(5:8))
ENDDO
gene%nmodes = 0
DO i = 1, ngenes
   IF (probdiff(i) < pct5_threshold ) gene(i)%nmodes = 2
   IF (probdiff(i) > pct95_threshold ) gene(i)%nmodes = 1
ENDDO


DO i = 1, ngenes
!  WRITE(90,fmat5)gene(i)%id,gene(i)%shape,gene(i)%alternative,gene(i)%nmodes
   WRITE(90,fmat5)gene(i)%id,gene(i)%shape
ENDDO
   

contains

function mean(x)
    implicit none
    real(8):: x(:),mean
    mean=sum(x)/size(x)
end function

function sd(x)
    implicit none
    real(8):: x(:),varx,sd
    varx=(sum(x**2)-sum(x)**2/size(x) )/(size(x)-1)
    sd=sqrt(varx)
end function


END PROGRAM CoExprDistrib

!#########################################################################
!#########################################################################
FUNCTION corr(x,y,n) RESULT(f_val)
IMPLICIT NONE

INTEGER, INTENT(IN) :: n
double precision, INTENT(IN) :: x(n),y(n)
double precision :: f_val

double precision :: s1, ss1, s2, ss2, ss12, var1, var2, num, den
INTEGER :: i,j

s1=0.0; ss1=0.0; s2=0.0; ss2=0.0; ss12=0.0

DO i = 1, n
   s1 = s1 + x(i)
   ss1 = ss1 + x(i)*x(i)
   s2 = s2 + y(i)
   ss2 = ss2 + y(i)*y(i)
   ss12 = ss12 + x(i)*y(i)
ENDDO
var1 = (ss1 - (s1*s1)/n) / (n-1)
var2 = (ss2 - (s2*s2)/n) / (n-1)
num = (ss12 - (s1*s2)/n) / (n-1)
den = SQRT(var1*var2)

f_val = num / den
if(f_val==0) then 
    print *,num,den,num/den
endif
END FUNCTION corr


subroutine number_of_fields(unit,n)
    ! guess number of fields in the 1st line of one file
 integer unit,n,i
 character(100000):: line
 character(1)::last
 last=' '
 n=0
 read(unit,'(a)')line
 do i=1,len_trim(line)
    if (line(i:i)/=' ' .and. last==' ') then
        n=n+1
    endif
    last=line(i:i)
 enddo
 !write (*,*) n
 rewind(unit)
end subroutine 

!#########################################################################
recursive subroutine quicksort(a, first, last)
  implicit none
  real*8  a(*), x, t
  integer first, last
  integer i, j

  x = a( (first+last) / 2 )
  i = first
  j = last
  do
     do while (a(i) < x)
        i=i+1
     end do
     do while (x < a(j))
        j=j-1
     end do
     if (i >= j) exit
     t = a(i);  a(i) = a(j);  a(j) = t
     i=i+1
     j=j-1
  end do
  if (first < i-1) call quicksort(a, first, i-1)
  if (j+1 < last)  call quicksort(a, j+1, last)
end subroutine quicksort
